#!/usr/bin/env python
"""
_WorkUnit_

Data object that contains details for a single work unit

"""

from __future__ import absolute_import, division, print_function

from future.utils import listitems

import sys
import hashlib
import time
from functools import total_ordering

from Utils.Utilities import encodeUnicodeToBytes
from WMCore.DataStructs.WMObject import WMObject


@total_ordering
class WorkUnit(WMObject, dict):
    """
    _WorkUnit_
    Data object that contains details for a single work unit
    corresponding to tables workunit and frl_workunit_assoc
    """

    fieldsToCopy = ['taskid', 'retry_count', 'last_unit_count', 'last_submit_time', 'status', 'firstevent',
                    'lastevent', 'fileid']
    fieldsForInfo = fieldsToCopy + ['run_lumi']

    def __init__(self, taskID=None, retryCount=0, lastUnitCount=None, lastSubmitTime=int(time.time()),
                 status=0, firstEvent=1, lastEvent=sys.maxsize, fileid=None, runLumi=None):
        super(WorkUnit, self).__init__(self)
        self.setdefault('taskid', taskID)
        self.setdefault('retry_count', retryCount)
        self.setdefault('last_unit_count', lastUnitCount)
        self.setdefault('last_submit_time', lastSubmitTime)
        self.setdefault('status', status)

        self.setdefault('firstevent', firstEvent)
        self.setdefault('lastevent', lastEvent)
        self.setdefault('fileid', fileid)
        self.setdefault('run_lumi', runLumi)

    def __lt__(self, rhs):
        """
        Compare work units in task id, run, lumi, first event, last event
        """

        if self['taskid'] != rhs['taskid']:
            return self['taskid'] < rhs['taskid']
        if self['run_lumi'].run != rhs['run_lumi'].run:
            return self['run_lumi'].run < rhs['run_lumi'].run
        if self['run_lumi'].lumis != rhs['run_lumi'].lumis:
            return self['run_lumi'].lumis < rhs['run_lumi'].lumis
        if self['first_event'] != rhs['first_event']:
            return self['first_event'] < rhs['first_event']
        return self['last_event'] < rhs['last_event']

    def __eq__(self, rhs):
        """
        Work unit is equal if it has the same task, run, and lumi
        """

        return (self['taskid'] == rhs['taskid'] and self['run_lumi'].run == self['run_lumi'].run and
                self['run_lumi'].lumis == self['run_lumi'].lumis and self['firstevent'] == rhs['firstevent'] and
                self['lastevent'] == rhs['lastevent'])

    def __hash__(self):
        """
        Hash function for this dict.
        """
        # Generate an immutable sorted string representing this object
        # NOTE: the run object needs to be hashed
        immutableSelf = []
        for keyName in sorted(self):
            if keyName == "run_lumi":
                immutableSelf.append((keyName, hash(self[keyName])))
            else:
                immutableSelf.append((keyName, self[keyName]))
        hashValue = hashlib.sha1(encodeUnicodeToBytes(str(immutableSelf)))
        return int(hashValue.hexdigest()[:15], 16)

    def json(self, thunker=None):
        """
        _json_

        Serialize the object.  Only copy select fields and construct one new field.
        """

        jsonDict = {k: self[k] for k in WorkUnit.fieldsToCopy}
        jsonDict["run_lumi"] = {"run_number": self['run_lumi'].run, "lumis": self['run_lumi'].lumis}

        return jsonDict

    def __to_json__(self, thunker=None):
        """
        __to_json__

        This is the standard way we jsonize other objects.
        Included here so we have a uniform method.
        """
        return self.json(thunker)

    def getInfo(self):
        """
        Returns: tuple of parameters for the work unit
        """
        return tuple(self[x] for x in WorkUnit.fieldsForInfo)



#!/usr/bin/env python
"""
Script to retrieve the request status and evaluate the workflow
completion, so the ration of input lumis vs output lumis.
"""
from __future__ import print_function, division

import argparse
import json
import os
import pwd
import sys

try:
    # python2
    import urllib2
    HTTPError, URLError = urllib2.HTTPError, urllib2.URLError
    build_opener = urllib2.build_opener
    HTTPSHandler = urllib2.HTTPSHandler
    import urllib
    urlencode = urllib.urlencode
    import httplib
    HTTPSConnection = httplib.HTTPSConnection
except ImportError:
    # python3
    import urllib.error
    HTTPError, URLError = urllib.error.HTTPError, urllib.error.URLError
    import urllib.request
    build_opener = urllib.request.build_opener
    HTTPSHandler = urllib.request.HTTPSHandler
    import urllib.parse
    urlencode = urllib.parse.urlencode
    import http.client
    HTTPSConnection = http.client.HTTPSConnection

# ID for the User-Agent
CLIENT_ID = 'workflowCompletion::python/%s.%s' % sys.version_info[:2]


class HTTPSClientAuthHandler(HTTPSHandler):
    """
    Basic HTTPS class
    """

    def __init__(self, key, cert):
        HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert

    def https_open(self, req):
        # Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=290):
        return HTTPSConnection(host, key_file=self.key, cert_file=self.cert)


def getX509():
    "Helper function to get x509 from env or tmp file"
    certFile = os.environ.get('X509_USER_CERT', '')
    keyFile = os.environ.get('X509_USER_KEY', '')
    if certFile and keyFile:
        return certFile, keyFile

    proxy = os.environ.get('X509_USER_PROXY', '')
    if not proxy:
        proxy = '/tmp/x509up_u%s' % pwd.getpwuid(os.getuid()).pw_uid
        if not os.path.isfile(proxy):
            return '', ''
    return proxy, proxy


def getContent(url, params=None):
    certFile, keyFile = getX509()
    client = '%s (%s)' % (CLIENT_ID, os.environ.get('USER', ''))
    handler = HTTPSClientAuthHandler(keyFile, certFile)
    opener = build_opener(handler)
    opener.addheaders = [("User-Agent", client),
                         ("Accept", "application/json")]

    try:
        response = opener.open(url, params)
        output = response.read()
    except HTTPError as e:
        print("The server couldn't fulfill the request at %s" % url)
        print("Error: {}".format(e))
        output = '{}'
        # sys.exit(1)
    except URLError as e:
        print('Failed to reach server at %s' % url)
        print('Reason: ', e.reason)
        sys.exit(2)
    return output


def handleReqMgr(reqName, reqmgrUrl):
    """
    Return the list of output datasets
    """
    urn = reqmgrUrl + "/reqmgr2/data/request/" + reqName
    reqmgrOut = json.loads(getContent(urn))['result'][0][reqName]

    if reqmgrOut['RequestStatus'] in ['assignment-approved', 'assigned', 'staging', 'staged']:
        print("Workflow %s in status: %s , skipping!\n" % (reqName, reqmgrOut['RequestStatus']))
        return None, None

    if 'InputDataset' in reqmgrOut:
        inputData = reqmgrOut['InputDataset']
    elif 'Task1' in reqmgrOut and 'InputDataset' in reqmgrOut['Task1']:
        inputData = reqmgrOut['Task1']['InputDataset']
    elif 'Step1' in reqmgrOut and 'InputDataset' in reqmgrOut['Step1']:
        inputData = reqmgrOut['Step1']['InputDataset']
    else:
        inputData = None

    print("==> %s\t(status: %s)" % (reqName, reqmgrOut['RequestStatus']))
    print("InputDataset:\n    %s (total input lumis: %s)" % (inputData, reqmgrOut['TotalInputLumis']))
    return reqmgrOut['TotalInputLumis'], reqmgrOut['OutputDatasets']


def handleDBS(reqmgrOutDsets, cmswebUrl):
    """
    Get total number of lumi sections in each dataset
    """
    if 'testbed' in cmswebUrl:
        dbsUrl = cmswebUrl + "/dbs/int/global/DBSReader"
    else:
        dbsUrl = cmswebUrl + "/dbs/prod/global/DBSReader"

    dbsOutput = {}
    for dataset in reqmgrOutDsets:
        fullUrl = dbsUrl + "filesummaries?" + urlencode({'dataset': dataset})
        data = json.loads(getContent(fullUrl))
        if data:
            dbsOutput[dataset] = data[0]['num_lumi']
        else:
            dbsOutput[dataset] = 0

    return dbsOutput


def main():
    """
    Requirements: you need to have your proxy and proper x509 environment
     variables set.

    Receive a workflow name in order to fetch the following information:
     - from couchdb: gets the output datasets and spec file
     - from phedex: gets dataset, block, files information
     - from dbs: gets dataset, block and files information
    """
    parser = argparse.ArgumentParser(description="Validate workflow input, output and config")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-w', '--workflow', help='A single workflow name')
    group.add_argument('-i', '--inputFile', help='Plain text file containing request names (one per line)')
    parser.add_argument('-c', '--cms', help='CMSWEB url to talk to DBS/PhEDEx. E.g: cmsweb-testbed.cern.ch')
    parser.add_argument('-r', '--reqmgr', help='Request Manager URL. Example: cmsweb-testbed.cern.ch')
    args = parser.parse_args()

    if args.workflow:
        listRequests = [args.workflow]
    elif args.inputFile:
        with open(args.inputFile, 'r') as f:
            listRequests = [req.rstrip('\n') for req in f.readlines()]
    else:
        parser.error("You must provide either a workflow name or an input file name.")
        sys.exit(3)

    cmswebUrl = "https://" + args.cms if args.cms else "https://cmsweb-prod.cern.ch"
    reqmgrUrl = "https://" + args.reqmgr if args.reqmgr else "https://cmsweb.cern.ch"

    for reqName in listRequests:
        ### Retrieve and process ReqMgr information
        inputLumis, reqmgrOutDsets = handleReqMgr(reqName, reqmgrUrl)
        if reqmgrOutDsets is None:
            continue

        ### Retrieve and process DBS information
        dbsInfo = handleDBS(reqmgrOutDsets, cmswebUrl)
        print("OutputDatasets:")
        for dset in reqmgrOutDsets:
            print("    %s (lumis: %s, lumi completion: %s)" % (dset, dbsInfo[dset], dbsInfo[dset] / inputLumis))
        print("")

    sys.exit(0)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Script to create a dedicated DAOFactory and execute a single DAO.

This script is intended to be used by developers and experts only.
The execution environment should be under any of the WMAgents we have. And the full set
of agent management and initialisation scripts need to be sourced in advance i.e.:

Usage:

      For production:
           source  /data/admin/wmagent/env.sh
           source  /data/srv/wmagent/current/apps/wmagent/etc/profile.d/init.sh
           python3 ExecuteDAO.py [--options] -- [SQL Query Arguments]

      For tier0:
           source  /data/tier0/admin/env.sh
           source  /data/tier0/srv/wmagent/current/apps/t0/etc/profile.d/init.sh
           python3 ExecuteDAO.py [--options] -- [SQL Query Arguments]

"""

import sys
import os
import re
import ast
import threading
import logging
import argparse
from pprint import pformat

from WMCore.DAOFactory import DAOFactory
from WMCore.WMInit import WMInit
from WMCore.Agent.Configuration import Configuration, loadConfigurationFile


def parseArgs():
    """
    Generic Argument Parser function
    """
    parser = argparse.ArgumentParser(
        prog='ExecuteDAO',
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__)

    parser.add_argument('-c', '--config', required=True,
                        help="""\
                        The path to WMAGENT_CONFIG to be used, e.g.
                        for production: /data/srv/wmagent/current/config/wmagent/config.py
                        for tier0:      /data/tier0/srv/wmagent/current/config/tier0/config.py""")
    parser.add_argument('-p', '--package', required=True,
                        help="""\
                        The package from which the DAO factory to be created for this execution, e.g. WMCore.WMBS or WMComponent.DBS3Buffer""")
    parser.add_argument('-m', '--module', required=True,
                        help="""\
                        The DAO Module to be executed, e.g. Workflow.GetDeletableWorkflows or CountUndeletedBlocksByWorkflow""")
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help="""\
                        Set logging to debug mode.""")
    parser.add_argument('--dryRun', action='store_true', default=False,
                        help="""\
                        Simulation mode only""")
    parser.add_argument('-s', '--sqlKwArgs', default={},
                        help="""\
                        Named paramaters to be forwarded to the DAO execute method and used as SQL arguments in the query.
                        Should be formatted as a dictionary e.g.:
                        -s "{'workflowName': name, injected: True}"
                        """)
    parser.add_argument('sqlArgs', nargs=argparse.REMAINDER, default=(),
                        help="""\
                        -- Positional parameters to be forwarded to the DAO execute method and used as SQL arguments in the query.""")

    args = parser.parse_args()

    return args


def loggerSetup(logLevel=logging.INFO):
    """
    Return a logger which writes everything to stdout.
    """
    logger = logging.getLogger()
    outHandler = logging.StreamHandler(sys.stdout)
    outHandler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(module)s: %(message)s"))
    outHandler.setLevel(logLevel)
    if logger.handlers:
        logger.handlers.clear()
    logger.addHandler(outHandler)
    logger.setLevel(logLevel)
    return logger


def getBackendFromDbURL(dburl):
    """
    Auxiliary function for determining sql dialect from a connection Url
    :param dbUrl: The connection Url to be parsed.
    :return: A string pointing to the correct dialect.
    """
    dialectPart = dburl.split(":")[0]
    if dialectPart == 'mysql':
        return 'MySQL'
    elif dialectPart == 'oracle':
        return 'Oracle'
    else:
        raise RuntimeError("Unrecognized dialect %s" % dialectPart)


class ExecuteDAO():
    """
    A generic class to create the DAO Factory and execute the DAO module.
    """
    def __init__(self, logger=None, configFile=None,
                 connectUrl=None, dialect=None, socket=None,
                 package=None, daoModule=None):
        """
        __init__
        The ExecuteDAO constructor method.
        :param logger: The logger instance.
        :param package: The Package from which the DAO factory to be initialised.
        :param configFile: Path to WMAgent configuration file.
        :param connectUrl: Database connection URL (overwrites the connectUrl param from configFile if both present)
        :param dialect: Database connection URL (overwrites the dialect param from configFile if both present)
        :param socket: Database connection URL (overwrites the socket param from configFile if both present)
        :param module: The DAO module to be executed.
        """
        # Get the current thread:
        myThread = threading.currentThread()

        # Create default WMCore Init thread and configs:
        self.init = WMInit()

        if logger is None:
            self.init.setLogging()
            self.logger = logging.getLogger()
        else:
            self.logger = logger

        if configFile is not None:
            self.logger.info("Loading configFile: %s", configFile)
            config = loadConfigurationFile(configFile)
        else:
            config = Configuration()

        # Overwrite database config parameters from configFile if present as init arguments:
        config.section_("CoreDatabase")
        if connectUrl is not None:
            config.CoreDatabase.connectUrl = connectUrl

        if dialect is not None:
            config.CoreDatabase.dialect = dialect

        if socket is not None:
            config.CoreDatabase.socket = socket

        # If still no proper database connection parameters provided,
        # last resort - try fetching them from the environment:
        if getattr(config.CoreDatabase, "connectUrl", None) is None and os.getenv('DATABASE', None):
            config.CoreDatabase.connectUrl = os.getenv('DATABASE')
            config.CoreDatabase.dialect = getBackendFromDbURL(os.getenv("DATABASE"))
            config.CoreDatabase.socket = os.getenv("DBSOCK", None)

        # check if all database connection parameters are provided:
        if not all([getattr(config.CoreDatabase, "connectUrl", None),
                    getattr(config.CoreDatabase, "dialect", None),
                    getattr(config.CoreDatabase, "socket", None)]):
            raise RuntimeError("You must set proper DATABASE parameters: connectUrl, dialect, socket!")

        # Connecting to database:
        self.init.setDatabaseConnection(config.CoreDatabase.connectUrl,
                                        config.CoreDatabase.dialect,
                                        config.CoreDatabase.socket)

        self.dbi = myThread.dbi
        self.package = package
        self.daoModule = daoModule

        # Avoid any name that starts with _
        self.sqlRegEx = re.compile("^(?!_.*)", re.IGNORECASE)

        self.daoFactory = DAOFactory(package=package,
                                     logger=self.logger,
                                     dbinterface=self.dbi)
        self.logger.info("DAO Factory initialised from package: %s", self.package)

        self.dao = self.daoFactory(classname=daoModule)
        self.logger.info("DAO Module initialised as: %s", self.daoModule)

    def __call__(self, *sqlArgs, dryRun=False, daoHelp=False, **sqlKwArgs):
        """
        __call__
        The ExecuteDAO call method. This is the method to forward all provided
        arguments to the execute method of the DAO and return the result from the query
        :param dryRun:      Bool flag to indicate dryrun method
        :param *sqlArgs:    All positional arguments to be forwarded to the DAO's execute method.
        :param **sqlKwArgs: All named arguments to be forwarded to the DAO's execute method.
        :return:            The result from the DAO. Depending on the DAO itself it Could be one of:
                            * A dictionary
                            * A list
                            * A generator
        """
        if dryRun:
            results = []
            if daoHelp:
                self.getHelp()
            self.logger.info("DAO SQL queries to be executed:")
            sqlQueries = self.getSqlQuery()
            for sqlName, sqlStr in sqlQueries.items():
                msg = "\n----------------------------------------------------------------------\n"
                msg += "%s: %s"
                msg += "\n----------------------------------------------------------------------\n"
                self.logger.info(msg, sqlName, sqlStr)
            self.logger.info("DAO SQL arguments provided:\n%s, %s", pformat(sqlArgs), pformat(sqlKwArgs))
        else:
            results = self.dao.execute(*sqlArgs, **sqlKwArgs)
            self.logger.info("DAO Results:\n%s", pformat(results if isinstance(results, dict) else list(results)))
        return results

    def getSqlQuery(self):
        """
        A simple method to inspect all DAO object attributes and accumulate any sql query it finds in a simple list
        :return: A list of all sql queries it finds in the object.
        """
        # NOTE: Use this method with caution because it may also return an object which is not an sql query.
        #       This may happen if there is a DAO attribute which is of type string and satisfies self.sqlRegEx
        sqlQueries = {}
        for attr in dir(self.dao):
            if self.sqlRegEx.match(attr) and isinstance(getattr(self.dao, attr), (str, bytes)):
                sqlQueries[attr] = getattr(self.dao, attr)
        return sqlQueries

    def getHelp(self):
        """
        A simple method to generate interactive DAO help message from the DAO source.
        """
        help(self.dao)


def strToDict(dString, logger=None):
    """
    A simple Function to parse a string and produce a dictionary out of it.
    :param dString: The dictionary string to be parsed. Possible formats are either a string
                    of multiple space separated named values of the form 'name=value':
                    or a srting fully defining the dictionary itself.
    :param logger:  A logger object to be used for Error message printout
    :return:        The constructed dictionary
    """
    logger = logger or logging.getLogger()
    result = ast.literal_eval(dString)
    if not isinstance(result, dict):
        logger.error("The Query named arguments need to be provided as a dictionary. WRONG option: %s", pformat(dString))
        raise TypeError(pformat(dString))
    return result


def main():
    """
    An Utility to construct a DAO Factory and execute the DAO requested.
    """
    args = parseArgs()

    if args.debug:
        logger = loggerSetup(logging.DEBUG)
    else:
        logger = loggerSetup()

    # Remove leading double slash if present:
    if args.sqlArgs and args.sqlArgs[0] == '--':
        args.sqlArgs = args.sqlArgs[1:]

    # Convert the positional arguments to a tuple:
    if not isinstance(args.sqlArgs, tuple):
        args.sqlArgs = tuple(args.sqlArgs)

    # Parse named arguments to a proper dictionary:
    if not isinstance(args.sqlKwArgs, dict):
        args.sqlKwArgs = strToDict(args.sqlKwArgs, logger=logger)

    # Try to set the wmagent configuration in the environment
    if 'WMAGENT_CONFIG' not in os.environ:
        os.environ['WMAGENT_CONFIG'] = args.config
    configFile = os.environ['WMAGENT_CONFIG']

    daoObject = ExecuteDAO(logger=logger, configFile=configFile,
                           package=args.package, daoModule=args.module)
    daoObject(*args.sqlArgs, dryRun=args.dryRun, daoHelp=True, **args.sqlKwArgs)


if __name__ == '__main__':
    main()
